#!/usr/bin/env python
# -*- coding: utf-8 -*-

from . import utils
from easyrocks import DB, WriteBatch
import logging


class Blockchain:

    # meta_ -> labels ro by the user
    # labels_ -> labels rw by the user
    # block_ -> blocks ro
    # hash_ -> blocks ro
    # height_ -> blocks ro

    @staticmethod
    def _get_block_key(height, prev_hash):
        return f'block_{utils.get_padded_int(height)}_{prev_hash}'

    def __init__(self, path='./chaindata', signature_checker=None, create_if_missing=False):
        self._db = DB(path=path, opts={'create_if_missing': create_if_missing})
        if signature_checker:
            self.signature_checker = signature_checker
        self._write_batch = None

    def put(self, key, value):
        self._put(f'label_{key}', value)

    def get(self, key):
        return self._db.get(f'label_{key}')

    def add_block(self, block_dict):
        self.begin_write_batch()
        height = self._increment_height()
        if self.get_block_by_hash(block_dict['hash']):
            raise PermissionError
        if not self._prev_hash_is_correct(height, block_dict['prev_hash']):
            raise PermissionError
        if not self._signature_is_valid(block_dict):
            raise PermissionError
        key = self._get_block_key(height, block_dict['hash'])
        self._put(f'block_{key}', block_dict)  # height, block_hash -> block_dict
        self._put(f'height_{height}', block_dict['hash'])  # height -> block_hash
        self._put(f"hash_{block_dict['hash']}", height)  # block_hash -> height
        self.commit_write_batch()

    def get_block_by_height(self, height):
        block_hash = self._db.get(f'height_{height}')
        return self._get_block(height, block_hash)

    def get_block_by_hash(self, block_hash):
        height = self._db.get(f'hash_{block_hash}')
        return self._get_block(height, block_hash)

    def get_block_hash_by_height(self, height):
        return self._db.get(f'height_{height}')

    def get_block_height_by_hash(self, block_hash):
        return self._db.get(f'hash_{block_hash}')

    def get_height(self):
        height = self.get_meta('height')
        return height

    def get_meta(self, key):
        return self._db.get(f'meta_{key}')

    def print_chain(self):
        for key, value in self._db.scan(prefix='block_'):
            height = key.split('_')[1]
            block_dict = utils.dump_pretty_dict(utils.load_dict(value))
            print(f'Height: {height}')
            print(block_dict)
            print()

    def print_meta(self):
        for key, value in self._db.scan(prefix='label_'):
            print(f'key: {key}, value: {value}')

    def begin_write_batch(self):
        if self._write_batch is None:
            self._write_batch = WriteBatch()

    def commit_write_batch(self):
        if self._write_batch is not None:
            self._db.commit(self._write_batch)
            self._write_batch = None

    def _get_block(self, height, block_hash):
        block_key = self._get_block_key(height, block_hash)

        block = self._db.get(block_key)
        if block is not None:
            return utils.load_dict(block)

    def _can_put(self, key):
        prefix = key.split('_')[0]

        if self._db.get(key) is None \
        or prefix == 'label' \
        or prefix == 'meta':
            return True
        else:
            return False

    def _put(self, key, value):
        if not self._can_put(key):
            raise PermissionError
        self._db.put(key, value, write_batch=self._write_batch)

    def _increment_height(self):
        height = self.get_height()
        if height is None:
            height = 0
        else:
            height += 1
        self._put('meta_height', height)
        return height

    def _prev_hash_is_correct(self, current_height, prev_hash):
        if self.get_height() is None:
            return True

        prev_height = self.get_block_height_by_hash(prev_hash)
        if prev_height != current_height - 1:
            return False

        block_dict = self.get_block_by_hash(prev_hash)
        if not block_dict:
            return False

        if block_dict['hash'] == prev_hash:
            return True

        return False

    def _signature_is_valid(self, block_dict):
        if not hasattr(self, 'signature_checker'):
            return True

        if not 'signature' in block_dict:
            return False
        return self.signature_checker(block_dict['signature'])